"""
# API keys and configuration

## Setting keys

To check whether appropriate api keys have been set:

```python
from trulens.core.utils.keys import check_keys

check_keys(
    "OPENAI_API_KEY",
    "HUGGINGFACE_API_KEY"
)
```

Alternatively you can set using `check_or_set_keys`:

```python
from trulens.core.utils.keys import check_or_set_keys

check_or_set_keys(
    OPENAI_API_KEY="to fill in",
    HUGGINGFACE_API_KEY="to fill in"
)
```

This line checks that you have the requisite api keys set before continuing the
notebook. They do not need to be provided, however, right on this line. There
are several ways to make sure this check passes:

- *Explicit* -- Explicitly provide key values to `check_keys`.

- *Python* -- Define variables before this check like this:

```python
OPENAI_API_KEY="something"
```

- *Environment* -- Set them in your environment variable. They should be visible when you execute:

```python
import os
print(os.environ)
```

- *.env* -- Set them in a .env file in the same folder as the example notebook or one of
  its parent folders. An example of a .env file is found in
  `trulens/trulens/env.example` .

- *Endpoint class* For some keys, set them as arguments to trulens endpoint class that
  manages the endpoint. For example, with `openai`, do this ahead of the
  `check_keys` check:

```python
from trulens.providers.openai import OpenAIEndpoint
openai_endpoint = OpenAIEndpoint(api_key="something")
```

- *Provider class* For some keys, set them as arguments to trulens feedback
  collection ("provider") class that makes use of the relevant endpoint. For
  example, with `openai`, do this ahead of the `check_keys` check:

```python
from trulens.providers.openai import OpenAI
openai_feedbacks = OpenAI(api_key="something")
```

In the last two cases, please note that the settings are global. Even if you
create multiple OpenAI or OpenAIEndpoint objects, they will share the
configuration of keys (and other openai attributes).

## Other API attributes

Some providers may require additional configuration attributes beyond api key.
For example, `openai` usage via azure require special keys. To set those, you
should use the 3rd party class method of configuration. For example with
`openai`:

```python
import openai

openai.api_type = "azure"
openai.api_key = "..."
openai.api_base = "https://example-endpoint.openai.azure.com"
openai.api_version = "2023-05-15"  # subject to change
# See https://learn.microsoft.com/en-us/azure/cognitive-services/openai/how-to/switching-endpoints .
```

Our example notebooks will only check that the api_key is set but will make use
of the configured openai object as needed to compute feedback.
"""

from collections import defaultdict
import logging
import os
from pathlib import Path
import re
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union

import dotenv
from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils

logger = logging.getLogger(__name__)

# Keep track of values that should not be shown in UI (or added to DB). This set
# is only used for cases where the name/key for a field is not useful to
# determine whether it should be redacted.
values_to_redact: Set[str] = set()

# Regex of keys (into dict/json) that should be redacted.
RE_KEY_TO_REDACT: re.Pattern = re.compile(
    "|".join([
        r"api_key",
        # Covers OpenAI, Cohere, Anthropic class key 'api_key'
        r".+_api_key",
        # Covers langchain llm attributes for keys such as 'openai_api_key'.
        # r'token',
        # Would cover bard unofficial api field "token" but this is a
        # bit too general of a key; TODO: need another solution to redact.
        r".+_API_KEY",
        r".+_API_TOKEN",
        # Covers env vars ending in "_API_KEY", including openai, cohere, anthropic,
        # bard
        r"KAGGLE_KEY",
        r"SLACK_(TOKEN|SIGNING_SECRET)",
        # Covers slack-related keys.
        r".*PASSWORD.*",
        # Cover user credentials for i.e. Snowflake connection session
    ])
)

# Env vars not covered as they are assumed non-sensitive:
# - PINECONE_ENV, e.g. "us-west1-gcp-free"
# - KAGGLE_USER

# Keys not covered that might be sensitive:
# - "token" - i.e. bard-api Bard.token, slack api's -- name collision with
#   "token" as in the basic building block of text.

# TODO: Some method for letting users add more things to redact.

# The replacement value for redacted values.
REDACTED_VALUE = "__tru_redacted"

# Treat these value as not valid keys. Use any as a templates to suggest a user
# fills in the key.
TEMPLATE_VALUES = {"to fill in"}

global cohere_agent
cohere_agent = None


def should_redact_key(k: Optional[str]) -> bool:
    return isinstance(k, str) and RE_KEY_TO_REDACT.fullmatch(k)


def should_redact_value(v: Union[Any, str]) -> bool:
    return isinstance(v, str) and v in values_to_redact


def redact_value(
    v: Union[str, Any], k: Optional[str] = None
) -> Union[str, Any]:
    """
    Determine whether the given value `v` should be redacted and redact it if
    so. If its key `k` (in a dict/json-like) is given, uses the key name to
    determine whether redaction is appropriate. If key `k` is not given, only
    redacts if `v` is a string and identical to one of the keys ingested using
    `setup_keys`.
    """

    if should_redact_key(k) or should_redact_value(v):
        return REDACTED_VALUE
    else:
        return v


def get_config_file() -> Optional[Path]:
    """
    Looks for a .env file in current folder or its parents. Returns Path of
    found .env or None if not found.
    """
    for path in [Path().cwd(), *Path.cwd().parents]:
        file = path / ".env"
        if file.exists():
            return file

    return None


def get_config() -> Tuple[Optional[Path], Optional[dict]]:
    config_file = get_config_file()
    if config_file is None:
        logger.warning(
            f"No .env found in {Path.cwd()} or its parents. "
            "You may need to specify secret keys in another manner."
        )
        return None, None
    else:
        return config_file, dotenv.dotenv_values(config_file)


def get_huggingface_headers() -> Dict[str, str]:
    HUGGINGFACE_HEADERS = {
        "Authorization": f"Bearer {os.environ['HUGGINGFACE_API_KEY']}"
    }
    return HUGGINGFACE_HEADERS


def _value_is_set(v: str) -> bool:
    return not (v is None or v in TEMPLATE_VALUES or v == "")


class BaseKeyError(RuntimeError):
    def __init__(self, *args: Iterable[Any]):
        super().__init__(*args)


class ApiKeyError(BaseKeyError):
    def __init__(self, *args: Iterable[Any], key: str):
        super().__init__(*args)
        self.key = key


class MissingKeyError(BaseKeyError):
    def __init__(self, *args: Iterable[Any], missing_keys: Iterable[str]):
        if not args:
            args = (f"Missing key(s): {', '.join(missing_keys)}",)
        super().__init__(*args)
        self.missing_keys = missing_keys


def _check_key(
    k: str, v: Optional[str] = None, silent: bool = False, warn: bool = False
) -> bool:
    """
    Check that the given `k` is an env var with a value that indicates a valid
    api key or secret.  If `v` is provided, checks that instead. If value
    indicates the key is not set, raises an informative error telling the user
    options on how to set that key. If `silent` is set, nothing will be printed.
    If `warn` is set, will log warning to logger and not throw an exception.
    Returns True if key is set. Silent disable warning logging.
    """

    v = v or os.environ.get(k)

    if not v or not _value_is_set(v):
        msg = f"""Key {k} needs to be set; please provide it in one of these ways:

  - in a variable {k} prior to this check,
  - in your variable environment,
  - in a .env file in {Path.cwd()} or its parents,
  - explicitly passed to function `check_or_set_keys` of `trulens.core.utils.keys`,
  - passed to the endpoint or feedback collection constructor that needs it (`trulens.providers.openai.endpoint.OpenAIEndpoint`, etc.), or
  - set in api utility class that expects it (i.e. `OpenAI(api_key=)`, etc.).

For the last two options, the name of the argument may differ from {k} (i.e. `OpenAI(api_key=)` for `OPENAI_API_KEY`).
"""
        if not silent:
            print(f"{text_utils.UNICODE_STOP} {msg}")
            if warn:
                logger.warning(msg)
        else:
            if warn:
                return False
            else:
                raise ApiKeyError(msg, key=k)

    return True


def _relative_path(path: Path, relative_to: Path) -> str:
    """
    Get the path `path` relative to path `relative_to` even if `relative_to` is
    not a prefix of `path`. Iteratively takes the parent of `relative_to` in
    that case until it becomes a prefix. Each parent is indicated by '..'.
    """

    parents = 0

    while True:
        try:
            return "".join(["../"] * parents) + str(
                path.relative_to(relative_to)
            )
        except Exception:
            parents += 1
            relative_to = relative_to.parent


def _collect_keys(*args: str, **kwargs: Dict[str, str]) -> Dict[str, str]:
    """
    Collect values for keys from all of the currently supported sources. This includes:

    - Using env variables.

    - Using python variables.

    - Explicitly passed to `check_or_set_keys`.

    - Using vars defined in a .env file in current folder or one of its parents.

    - With initialization of trulens Endpoint class that handles a 3rd party api.
    """

    ret = dict()

    config_file, config = get_config()

    globs = python_utils.caller_frame(offset=2).f_globals

    for k in list(args) + list(kwargs.keys()):
        valid_values = set()
        valid_sources = defaultdict(list)

        # Env vars. NOTE: Endpoint classes copy over relevant keys from 3rd party
        # classes (or provided explicitly to them) to var env.
        temp_v = os.environ.get(k)
        if temp_v and _value_is_set(temp_v):
            valid_sources[temp_v].append("environment")
            valid_values.add(temp_v)

        # Explicit.
        temp_v = kwargs.get(k)
        if temp_v and _value_is_set(temp_v):
            valid_sources[temp_v].append(
                "explicit value to `check_or_set_keys`"
            )
            valid_values.add(temp_v)

        # .env vars.
        if config is not None:
            temp_v = config.get(k)
            if temp_v and _value_is_set(temp_v):
                valid_sources[temp_v].append(f".env file at {config_file}")
                valid_values.add(temp_v)

        # Globals of caller.
        temp_v = globs.get(k)
        if temp_v and _value_is_set(temp_v):
            valid_sources[temp_v].append("python variable")
            valid_values.add(temp_v)

        if len(valid_values) == 0:
            ret[k] = None

        elif len(valid_values) > 1:
            warning = (
                f"More than one different value for key {k} has been found:\n\t"
            )
            warning += "\n\t".join(
                f"""value ending in {v[-1]} in {" and ".join(valid_sources[v])}"""
                for v in valid_values
            )
            warning += "\nUsing one arbitrarily."
            logger.warning(warning)

            ret[k] = list(valid_values)[0]
        else:
            v = list(valid_values)[0]
            print(
                f"{text_utils.UNICODE_CHECK} Key {k} set from {valid_sources[v][0]}"
                + (
                    " (same value found in "
                    + (" and ".join(valid_sources[v][1:]))
                    + ")"
                    if len(valid_sources[v]) > 1
                    else ""
                )
                + "."
            )

            ret[k] = v

    return ret


def check_keys(*keys: str) -> None:
    """
    Check that all keys named in `*args` are set as env vars. Will fail with a
    message on how to set missing key if one is missing. If all are provided
    somewhere, they will be set in the env var as the canonical location where
    we should expect them subsequently.

    Example:
        ```python
        from trulens.core.utils.keys import check_keys

        check_keys(
            "OPENAI_API_KEY",
            "HUGGINGFACE_API_KEY"
        )
        ```
    """

    kvals = _collect_keys(*keys)
    missing_keys = []
    for k in keys:
        v = kvals.get(k)
        _check_key(k, v=v)
        if v is not None:
            # Put value in redaction list.
            values_to_redact.add(v)
            os.environ[k] = v
        else:
            missing_keys.append(k)
    if missing_keys:
        raise MissingKeyError(missing_keys=missing_keys)


def check_or_set_keys(*args: str, **kwargs: Dict[str, str]) -> None:
    """
    Check various sources of api configuration values like secret keys and set
    env variables for each of them. We use env variables as the canonical
    storage of these keys, regardless of how they were specified. Values can
    also be specified explicitly to this method. Example:
    ```python
    from trulens.core.utils.keys import check_or_set_keys

    check_or_set_keys(
        OPENAI_API_KEY="to fill in",
        HUGGINGFACE_API_KEY="to fill in"
    )
    ```
    """

    kvals = _collect_keys(*args, **kwargs)
    for k in list(args) + list(kwargs.keys()):
        v = kvals.get(k)
        _check_key(k, v=v)
        if v:
            values_to_redact.add(v)
            os.environ[k] = v



"""Utilities related to guessing inputs and outputs of functions."""

from inspect import BoundArguments
from inspect import Signature
import logging
from typing import Any, Callable, Dict, Sequence

import pydantic
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)


def _extract_content(value, content_keys=["content"]):
    """
    Extracts the 'content' from various data types commonly used by libraries
    like OpenAI, Canopy, LiteLLM, etc. This method navigates nested data
    structures (pydantic models, dictionaries, lists) to retrieve the
    'content' field. If 'content' is not directly available, it attempts to
    extract from known structures like 'choices' in a ChatResponse. This
    standardizes extracting relevant text or data from complex API responses
    or internal data representations.

    Args:
        value: The input data to extract content from. Can be a pydantic
                model, dictionary, list, or basic data type.

    Returns:
        The extracted content, which may be a single value, a list of values,
        or a nested structure with content extracted from all levels.
    """
    if isinstance(value, pydantic.BaseModel):
        content = getattr(value, "content", None)
        if content is not None:
            return content

        # If 'content' is not found, check for 'choices' attribute which indicates a ChatResponse
        choices = getattr(value, "choices", None)
        if choices is not None:
            # Extract 'content' from the 'message' attribute of each _Choice in 'choices'
            return [_extract_content(choice.message) for choice in choices]

        # Recursively extract content from nested pydantic models
        try:
            return {
                k: _extract_content(v)
                if isinstance(v, (pydantic.BaseModel, dict, list))
                else v
                for k, v in value.model_dump().items()
            }
        except Exception as e:
            logger.warning(
                "Failed to extract content from pydantic model: %s", e
            )
            # Unsure what is best to do here. Lets just return the exception
            # so that it might reach the user if nothing else gets picked up
            # as main input/output.
            return str(e)

    elif isinstance(value, dict):
        # Check for 'content' key in the dictionary
        for key in content_keys:
            content = value.get(key)
            if content is not None:
                return content

        # Recursively extract content from nested dictionaries
        return {
            k: _extract_content(v) if isinstance(v, (dict, list)) else v
            for k, v in value.items()
        }

    elif isinstance(value, list):
        # Handle lists by extracting content from each item
        return [_extract_content(item) for item in value]

    else:
        return value


def main_input(func: Callable, sig: Signature, bindings: BoundArguments) -> str:
    """Determine (guess) the main input string for a main app call.

    Args:
        func: The main function we are targeting in this determination.

        sig: The signature of the above.

        bindings: The arguments to be passed to the function.

    Returns:
        The main input string.
    """

    if bindings is None:
        raise RuntimeError(
            f"Cannot determine main input of unbound call to {func}: {sig}."
        )

    # ignore self
    all_args = list(
        v for k, v in bindings.arguments.items() if k not in ["self", "_self"]
    )  # llama_index is using "_self" in some places

    # If there is only one string arg, it is a pretty good guess that it is
    # the main input.

    # if have only containers of length 1, find the innermost non-container
    focus = all_args

    while not isinstance(focus, serial_utils.JSON_BASES) and len(focus) == 1:
        focus = focus[0]
        focus = _extract_content(focus, content_keys=["content", "input"])

        if not isinstance(focus, Sequence):
            logger.warning("Focus %s is not a sequence.", focus)
            break

    if isinstance(focus, serial_utils.JSON_BASES):
        return str(focus)

    # Otherwise we are not sure.
    logger.warning(
        "Unsure what the main input string is for the call to %s with args %s.",
        python_utils.callable_name(func),
        bindings,
    )

    # After warning, just take the first item in each container until a
    # non-container is reached.
    focus = all_args
    while not isinstance(focus, serial_utils.JSON_BASES) and len(focus) >= 1:
        focus = focus[0]
        focus = _extract_content(focus)

        if not isinstance(focus, Sequence):
            logger.warning("Focus %s is not a sequence.", focus)
            break

    if isinstance(focus, serial_utils.JSON_BASES):
        return str(focus)

    logger.warning(
        "Could not determine main input/output of %s.", str(all_args)
    )

    return "TruLens: Could not determine main input from " + str(all_args)


def main_output(func: Callable, ret: Any) -> str:
    """Determine (guess) the "main output" string for a given main app call.

    This is for functions whose output is not a string.

    Args:
        func: The main function whose main output we are guessing.

        ret: The return value of the function.
    """

    if isinstance(ret, serial_utils.JSON_BASES):
        return str(ret)

    if isinstance(ret, Sequence) and all(isinstance(x, str) for x in ret):
        # Chunked/streamed outputs.
        return "".join(ret)

    # Use _extract_content to get the content out of the return value
    content = _extract_content(ret, content_keys=["content", "output"])

    if isinstance(content, str):
        return content

    if isinstance(content, float):
        return str(content)

    if isinstance(content, Dict):
        return str(next(iter(content.values()), ""))

    error_message = (
        f"Could not determine main output of {func.__name__}"
        f" from {python_utils.class_name(type(content))} value {content}."
    )

    if isinstance(content, Sequence):
        if len(content) > 0:
            return str(content[0])
        else:
            return error_message

    else:
        logger.warning(error_message)
        return (
            str(content)
            if content is not None
            else (f"TruLens: {error_message}")
        )



import inspect
import logging
import types
from types import TracebackType
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

from opentelemetry import trace
from opentelemetry.baggage import get_baggage
from opentelemetry.baggage import remove_baggage
from opentelemetry.baggage import set_baggage
import opentelemetry.context as context_api
from opentelemetry.trace.span import Span
from trulens.core.otel.function_call_context_manager import (
    create_function_call_context_manager,
)
from trulens.experimental.otel_tracing.core.session import TRULENS_SERVICE_NAME
from trulens.experimental.otel_tracing.core.span import Attributes
from trulens.experimental.otel_tracing.core.span import (
    set_function_call_attributes,
)
from trulens.experimental.otel_tracing.core.span import (
    set_general_span_attributes,
)
from trulens.experimental.otel_tracing.core.span import (
    set_record_root_span_attributes,
)
from trulens.experimental.otel_tracing.core.span import (
    set_user_defined_attributes,
)
from trulens.otel.semconv.constants import TRULENS_INSTRUMENT_WRAPPER_FLAG
from trulens.otel.semconv.constants import (
    TRULENS_RECORD_ROOT_INSTRUMENT_WRAPPER_FLAG,
)
from trulens.otel.semconv.trace import SpanAttributes
import wrapt

logger = logging.getLogger(__name__)


def _get_func_name(func: Callable) -> str:
    if (
        hasattr(func, "__module__")
        and func.__module__
        and hasattr(func, "__qualname__")
        and func.__qualname__
    ):
        return f"{func.__module__}.{func.__qualname__}"
    elif hasattr(func, "__qualname__") and func.__qualname__:
        return func.__qualname__
    else:
        return func.__name__


def _resolve_attributes(
    attributes: Attributes,
    ret: Optional[Any],
    exception: Optional[Exception],
    args: Sequence[Any],
    all_kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    if attributes is None:
        return {}
    if callable(attributes):
        return attributes(ret, exception, *args, **all_kwargs)
    if isinstance(attributes, dict):
        resolved = {}
        value_string_to_value = all_kwargs.copy()
        value_string_to_value["return"] = ret
        for k, v in attributes.items():
            resolved[k] = value_string_to_value[v]
        return resolved
    return attributes.copy()


def _set_span_attributes(
    span: Span,
    span_type: SpanAttributes.SpanType,
    span_name: str,
    func: Callable,
    func_exception: Optional[Exception],
    attributes: Attributes,
    instance: Any,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    ret: Any,
    only_set_user_defined_attributes: bool = False,
):
    # Determine args/kwargs to pass to the attributes
    # callable.
    sig = inspect.signature(func)
    bound_args = sig.bind_partial(*args, **kwargs).arguments
    all_kwargs = {**kwargs, **bound_args}
    if not only_set_user_defined_attributes:
        # Set general span attributes.
        span.set_attribute("name", span_name)
        set_general_span_attributes(span, span_type)
        # Set record root span attributes if necessary.
        if span_type == SpanAttributes.SpanType.RECORD_ROOT:
            set_record_root_span_attributes(
                span,
                func,
                args,
                kwargs,
                ret,
                func_exception,
            )
        # Set function call attributes.
        set_function_call_attributes(span, ret, func_exception, all_kwargs)
    # Resolve the attributes.
    if instance is not None:
        args_with_self_possibly = (instance,) + args
    else:
        args_with_self_possibly = args
    resolved_attributes = _resolve_attributes(
        attributes,
        ret,
        func_exception,
        args_with_self_possibly,
        all_kwargs,
    )
    if resolved_attributes:
        # Set the user-provided attributes.
        set_user_defined_attributes(
            span,
            span_type=span_type,
            attributes=resolved_attributes,
        )


def instrument(
    *,
    span_type: SpanAttributes.SpanType = SpanAttributes.SpanType.UNKNOWN,
    attributes: Attributes = None,
    must_be_first_wrapper: bool = False,
    **kwargs,
) -> Callable[[Callable], Any]:  # TODO(this_pr): get type!
    """
    Decorator for marking functions to be instrumented with OpenTelemetry
    tracing.

    span_type: Span type to be used for the span.
    attributes:
        A dictionary or a callable that returns a dictionary of attributes
        (i.e. a `typing.Dict[str, typing.Any]`) to be set on the span.
    must_be_first_wrapper:
        If this is True and the function is already wrapped with the TruLens
        decorator, then the function will not be wrapped again.
    """
    if attributes is None:
        attributes = {}
    is_record_root = span_type == SpanAttributes.SpanType.RECORD_ROOT
    is_app_specific_record_root = kwargs.get(
        "is_app_specific_record_root", False
    )
    create_new_span = kwargs.get("create_new_span", True)
    only_set_user_defined_attributes = kwargs.get(
        "only_set_user_defined_attributes", False
    )

    def inner_decorator(func: Callable):
        span_name = _get_func_name(func)

        @wrapt.decorator
        def sync_wrapper(func, instance, args, kwargs):
            ret = convert_to_generator(func, instance, args, kwargs)
            if next(ret) == "is_not_generator":
                res = next(ret)
                # Check that there are no more entries in the generator.
                valid = False
                try:
                    next(ret)
                except StopIteration:
                    valid = True
                if not valid:
                    raise ValueError("The generator is not empty!")
                ret = res
            return ret

        def convert_to_generator(func, instance, args, kwargs):
            with create_function_call_context_manager(
                create_new_span, span_name, is_record_root
            ) as span:
                ret = None
                func_exception: Optional[Exception] = None
                attributes_exception: Optional[Exception] = None
                # Run function.
                try:
                    result = func(*args, **kwargs)
                    if isinstance(result, types.GeneratorType):
                        yield "is_generator"
                        ret = []
                        for curr in result:
                            ret.append(curr)
                            yield curr
                    else:
                        yield "is_not_generator"
                        ret = result
                        yield ret
                except Exception as e:
                    # We want to get into the next clause to allow the users
                    # to still add attributes. It's on the user to deal with
                    # None as a return value.
                    func_exception = e
                finally:
                    # Set span attributes.
                    try:
                        _set_span_attributes(
                            span,
                            span_type,
                            span_name,
                            func,
                            func_exception,
                            attributes,
                            instance,
                            args,
                            kwargs,
                            ret,
                            only_set_user_defined_attributes=only_set_user_defined_attributes,
                        )
                    except Exception as e:
                        logger.error(f"Error setting attributes: {e}")
                        attributes_exception = e
                    # Raise any exceptions that occurred.
                    exception = func_exception or attributes_exception
                    if exception:
                        raise exception
                    return ret

        @wrapt.decorator
        async def async_wrapper(func, instance, args, kwargs):
            with create_function_call_context_manager(
                create_new_span, span_name, is_record_root
            ) as span:
                ret = None
                func_exception: Optional[Exception] = None
                attributes_exception: Optional[Exception] = None
                # Run function.
                try:
                    ret = await func(*args, **kwargs)
                except Exception as e:
                    # We want to get into the next clause to allow the users
                    # to still add attributes. It's on the user to deal with
                    # None as a return value.
                    func_exception = e
                # Set span attributes.
                try:
                    _set_span_attributes(
                        span,
                        span_type,
                        span_name,
                        func,
                        func_exception,
                        attributes,
                        instance,
                        args,
                        kwargs,
                        ret,
                    )
                except Exception as e:
                    logger.error(f"Error setting attributes: {e}")
                    attributes_exception = e
                # Raise any exceptions that occurred.
                exception = func_exception or attributes_exception
                if exception:
                    raise exception
            return ret

        @wrapt.decorator
        async def async_generator_wrapper(func, instance, args, kwargs):
            with create_function_call_context_manager(
                create_new_span, span_name, is_record_root
            ) as span:
                ret = None
                func_exception: Optional[Exception] = None
                attributes_exception: Optional[Exception] = None
                # Run function.
                try:
                    result = func(*args, **kwargs)
                    ret = []
                    async for curr in result:
                        ret.append(curr)
                        yield curr
                except Exception as e:
                    # We want to get into the next clause to allow the users
                    # to still add attributes. It's on the user to deal with
                    # None as a return value.
                    func_exception = e
                finally:
                    # Set span attributes.
                    try:
                        _set_span_attributes(
                            span,
                            span_type,
                            span_name,
                            func,
                            func_exception,
                            attributes,
                            instance,
                            args,
                            kwargs,
                            ret,
                        )
                    except Exception as e:
                        logger.error(f"Error setting attributes: {e}")
                        attributes_exception = e
                    # Raise any exceptions that occurred.
                    exception = func_exception or attributes_exception
                    if exception:
                        raise exception

        # Check if already wrapped if not allowing multiple wrappers.
        if must_be_first_wrapper:
            if hasattr(func, TRULENS_INSTRUMENT_WRAPPER_FLAG):
                return func
            curr = func
            while hasattr(curr, "__wrapped__"):
                curr = curr.__wrapped__
                if hasattr(curr, TRULENS_INSTRUMENT_WRAPPER_FLAG):
                    return func

        # Wrap.
        ret = None
        if inspect.isasyncgenfunction(func):
            ret = async_generator_wrapper(func)
        elif inspect.iscoroutinefunction(func):
            ret = async_wrapper(func)
        else:
            ret = sync_wrapper(func)
        ret.__dict__[TRULENS_INSTRUMENT_WRAPPER_FLAG] = True
        if is_record_root and not is_app_specific_record_root:
            ret.__dict__[TRULENS_RECORD_ROOT_INSTRUMENT_WRAPPER_FLAG] = True
        return ret

    return inner_decorator


def instrument_method(
    cls: type,
    method_name: str,
    span_type: SpanAttributes.SpanType = SpanAttributes.SpanType.UNKNOWN,
    attributes: Attributes = None,
    must_be_first_wrapper: bool = False,
) -> None:
    wrapper = instrument(
        span_type=span_type,
        attributes=attributes,
        must_be_first_wrapper=must_be_first_wrapper,
    )
    setattr(cls, method_name, wrapper(getattr(cls, method_name)))


def instrument_cost_computer(
    cls: type,
    method_name: str,
    attributes: Attributes,
) -> None:
    wrapper = instrument(
        attributes=attributes,
        create_new_span=False,
        only_set_user_defined_attributes=True,
    )
    setattr(cls, method_name, wrapper(getattr(cls, method_name)))


class OtelBaseRecordingContext:
    run_name: str
    """
    The name of the run that the recording context is currently processing.
    """

    input_id: str
    """
    The ID of the input that the recording context is currently processing.
    """

    tokens: List[object] = []
    """
    OTEL context tokens for the current context manager. These tokens are how the OTEL
    context api keeps track of what is changed in the context, and used to undo the changes.
    """

    context_keys_added: List[str] = []
    """
    Keys added to the OTEL context.
    """

    def __init__(
        self, *, app_name: str, app_version: str, run_name: str, input_id: str
    ):
        self.app_name = app_name
        self.app_version = app_version
        self.run_name = run_name
        self.input_id = input_id
        self.tokens = []
        self.context_keys_added = []
        self.span_context = None

    # Calling set_baggage does not actually add the baggage to the current context, but returns a new one
    # To avoid issues with remembering to add/remove the baggage, we attach it to the runtime context.
    def attach_to_context(self, key: str, value: object):
        if get_baggage(key) or value is None:
            return

        self.tokens.append(context_api.attach(set_baggage(key, value)))
        self.context_keys_added.append(key)

    async def __aenter__(self):
        return self.__enter__()

    def __exit__(self, exc_type, exc_value, exc_tb):
        # Exiting the span context before updating the context to ensure nothing
        # carries over unintentionally
        if self.span_context:
            # TODO[SNOW-1854360]: Add in feature function spans.
            self.span_context.__exit__(exc_type, exc_value, exc_tb)

        logger.debug("Exiting the OTEL app context.")

        # Clearing the context / baggage added.
        while self.context_keys_added:
            remove_baggage(self.context_keys_added.pop())

        while self.tokens:
            # Clearing the context once we're done with this root span.
            # See https://github.com/open-telemetry/opentelemetry-python/issues/2432#issuecomment-1593458684
            context_api.detach(self.tokens.pop())

    async def __aexit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        return self.__exit__(exc_type, exc_val, exc_tb)


class OtelRecordingContext(OtelBaseRecordingContext):
    def __init__(
        self,
        *,
        app_name: str,
        app_version: str,
        run_name: str,
        input_id: str,
        ground_truth_output: Optional[str] = None,
    ) -> None:
        super().__init__(
            app_name=app_name,
            app_version=app_version,
            run_name=run_name,
            input_id=input_id,
        )
        self.ground_truth_output = ground_truth_output

    # For use as a context manager.
    def __enter__(self) -> None:
        self.attach_to_context(SpanAttributes.APP_NAME, self.app_name)
        self.attach_to_context(SpanAttributes.APP_VERSION, self.app_version)

        self.attach_to_context(SpanAttributes.RUN_NAME, self.run_name)
        self.attach_to_context(SpanAttributes.INPUT_ID, self.input_id)
        self.attach_to_context(
            SpanAttributes.RECORD_ROOT.GROUND_TRUTH_OUTPUT,
            self.ground_truth_output,
        )


class OtelFeedbackComputationRecordingContext(OtelBaseRecordingContext):
    def __init__(self, *args, **kwargs):
        self.target_record_id = kwargs.pop("target_record_id")
        self.feedback_name = kwargs.pop("feedback_name")
        super().__init__(*args, **kwargs)

    # For use as a context manager.
    def __enter__(self) -> Span:
        tracer = trace.get_tracer_provider().get_tracer(TRULENS_SERVICE_NAME)

        self.attach_to_context(
            SpanAttributes.RECORD_ID, self.target_record_id
        )  # TODO(otel): Should we include this? It's automatically getting added to the span.
        self.attach_to_context(SpanAttributes.APP_NAME, self.app_name)
        self.attach_to_context(SpanAttributes.APP_VERSION, self.app_version)

        self.attach_to_context(SpanAttributes.RUN_NAME, self.run_name)
        self.attach_to_context(
            SpanAttributes.EVAL.TARGET_RECORD_ID, self.target_record_id
        )
        self.attach_to_context(SpanAttributes.INPUT_ID, self.input_id)
        self.attach_to_context(
            SpanAttributes.EVAL.METRIC_NAME, self.feedback_name
        )

        # Use start_as_current_span as a context manager
        self.span_context = tracer.start_as_current_span("eval_root")
        root_span = self.span_context.__enter__()
        root_span_id = str(root_span.get_span_context().span_id)

        self.attach_to_context(
            SpanAttributes.EVAL.EVAL_ROOT_ID,
            root_span_id,
        )

        # Set general span attributes
        root_span.set_attribute("name", "eval_root")
        set_general_span_attributes(
            root_span, SpanAttributes.SpanType.EVAL_ROOT
        )

        return root_span



import logging
from typing import Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter
from opentelemetry.sdk.trace.export import SpanExportResult
from trulens.core.database import connector as core_connector
from trulens.experimental.otel_tracing.core.exporter.utils import (
    check_if_trulens_span,
)
from trulens.experimental.otel_tracing.core.exporter.utils import (
    construct_event,
)

logger = logging.getLogger(__name__)


class TruLensOtelSpanExporter(SpanExporter):
    """
    Implementation of `SpanExporter` that flushes the spans in the TruLens session to the connector.
    """

    connector: core_connector.DBConnector
    """
    The database connector used to export the spans.
    """

    def __init__(self, connector: core_connector.DBConnector):
        self.connector = connector

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        trulens_spans = list(filter(check_if_trulens_span, spans))

        try:
            events = list(map(construct_event, trulens_spans))
            self.connector.add_events(events)

        except Exception as e:
            logger.error(
                f"Error exporting spans to the database: {e}",
            )
            return SpanExportResult.FAILURE

        return SpanExportResult.SUCCESS



"""Serializable feedback-related classes."""

from __future__ import annotations

import datetime
from enum import Enum
import logging
from pprint import pformat
from typing import (
    Any,
    ClassVar,
    Dict,
    Hashable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import pydantic
from trulens.core.schema import base as base_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils

T = TypeVar("T")

logger = logging.getLogger(__name__)


class FeedbackMode(str, Enum):
    """Mode of feedback evaluation.

    Specify this using the `feedback_mode` to [App][trulens.core.app.App]
    constructors.

    !!! Note
        This class extends [str][str] to allow users to compare its values with
        their string representations, i.e. in `if mode == "none": ...`. Internal
        uses should use the enum instances.
    """

    NONE = "none"
    """No evaluation will happen even if feedback functions are specified."""

    WITH_APP = "with_app"
    """Try to run feedback functions immediately and before app returns a
    record."""

    WITH_APP_THREAD = "with_app_thread"
    """Try to run feedback functions in the same process as the app but after
    it produces a record."""

    DEFERRED = "deferred"
    """Evaluate later via the process started by
    `TruSession.start_deferred_feedback_evaluator`."""


class FeedbackRunLocation(str, Enum):
    """Where the feedback evaluation takes place (e.g. locally, at a Snowflake server, etc)."""

    IN_APP = "in_app"
    """Run on the same process (or child process) of the app invocation."""

    SNOWFLAKE = "snowflake"
    """Run on a Snowflake server."""


class FeedbackResultStatus(str, Enum):
    """For deferred feedback evaluation, these values indicate status of
    evaluation.

    !!! Note
        This class extends [str][str] to allow users to compare its values with
        their string representations, i.e. in `if status == "done": ...`. Internal
        uses should use the enum instances.
    """

    NONE = "none"
    """Initial value is none."""

    RUNNING = "running"
    """Once queued/started, status is updated to "running"."""

    FAILED = "failed"
    """Run failed."""

    DONE = "done"
    """Run completed successfully."""

    SKIPPED = "skipped"
    """This feedback was skipped.

    This can be because because it had an `if_exists` selector and did not
    select anything or it has a selector that did not select anything the
    `on_missing` was set to warn or ignore.
    """


class FeedbackOnMissingParameters(str, Enum):
    """How to handle missing parameters in feedback function calls.

    This is specifically for the case were a feedback function has a selector
    that selects something that does not exist in a record/app.

    !!! Note
        This class extends [str][str] to allow users to compare its values with
        their string representations, i.e. in `if onmissing == "error": ...`.
        Internal uses should use the enum instances.
    """

    ERROR = "error"
    """Raise an error if a parameter is missing.

    The result status will be set to
    [FAILED][trulens.core.schema.feedback.FeedbackResultStatus.FAILED].
    """

    WARN = "warn"
    """Warn if a parameter is missing.

    The result status will be set to
    [SKIPPED][trulens.core.schema.feedback.FeedbackResultStatus.SKIPPED].
    """

    IGNORE = "ignore"
    """Do nothing.

    No warning or error message will be shown. The result status will be set to
    [SKIPPED][trulens.core.schema.feedback.FeedbackResultStatus.SKIPPED].
    """


class FeedbackCall(serial_utils.SerialModel):
    """Invocations of feedback function results in one of these instances.

    Note that a single `Feedback` instance might require more than one call.
    """

    args: Dict[str, Optional[serial_utils.JSON]] = pydantic.Field(
        default_factory=dict
    )
    """Arguments to the feedback function."""

    ret: Union[float, List[float], List[Tuple], List[Any]] = pydantic.Field(
        default=0.0
    )
    """Return value."""

    meta: Dict[str, Any] = pydantic.Field(default_factory=dict)
    """Any additional data a feedback function returns to display alongside its float result."""

    def __str__(self) -> str:
        out = ""
        tab = "  "
        for k, v in self.args.items():
            out += f"{tab}{k} = {v}\n"
        out += f"{tab}ret = {self.ret}\n"
        if self.meta:
            out += f"{tab}meta = \n{text_utils.retab(tab=tab * 2, s=pformat(self.meta))}\n"

        return out

    def __repr__(self) -> str:
        return str(self)


class FeedbackResult(serial_utils.SerialModel):
    """Feedback results for a single [Feedback][trulens.core.Feedback] instance.

    This might involve multiple feedback function calls. Typically you should
    not be constructing these objects yourself except for the cases where you'd
    like to log human feedback.

    Attributes:
        feedback_result_id: Unique identifier for this result.

        record_id: Record over which the feedback was evaluated.

        feedback_definition_id: The id of the
            [FeedbackDefinition][trulens.core.schema.feedback.FeedbackDefinition] which
            was evaluated to get this result.

        last_ts: Last timestamp involved in the evaluation.

        status: For deferred feedback evaluation, the
            status of the evaluation.

        cost: Cost of the evaluation.

        name: Given name of the feedback.

        calls: Individual feedback function invocations.

        result: Final result, potentially aggregating multiple calls.

        error: Error information if there was an error.

        multi_result: TBD
    """

    feedback_result_id: types_schema.FeedbackResultID

    # Record over which the feedback was evaluated.
    record_id: types_schema.RecordID

    # The `Feedback` / `FeedbackDefinition` which was evaluated to get this
    # result.
    feedback_definition_id: Optional[types_schema.FeedbackDefinitionID] = None

    # Last timestamp involved in the evaluation.
    last_ts: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.now
    )

    status: FeedbackResultStatus = FeedbackResultStatus.NONE
    """For deferred feedback evaluation, the status of the evaluation."""

    cost: base_schema.Cost = pydantic.Field(default_factory=base_schema.Cost)

    # Given name of the feedback.
    name: str

    # Individual feedback function invocations.
    calls: List[FeedbackCall] = []

    # Final result, potentially aggregating multiple calls.
    result: Optional[float] = None

    # Error information if there was an error.
    error: Optional[str] = None

    # TODO: doc
    multi_result: Optional[str] = None

    def __init__(
        self,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        **kwargs,
    ):
        super().__init__(feedback_result_id="temporary", **kwargs)

        if feedback_result_id is None:
            feedback_result_id = json_utils.obj_id_of_obj(
                self.model_dump(), prefix="feedback_result"
            )

        self.feedback_result_id = feedback_result_id

    def __str__(self):
        out = f"{self.name} ({self.status}) = {self.result}\n"
        for call in self.calls:
            out += pformat(call)

        return out

    def __repr__(self):
        return str(self)


class FeedbackCombinations(str, Enum):
    """How to collect arguments for feedback function calls.

    Note that this applies only to cases where selectors pick out more than one
    thing for feedback function arguments. This option is used for the field
    `combinations` of
    [FeedbackDefinition][trulens.core.schema.feedback.FeedbackDefinition] and can be
    specified with
    [Feedback.aggregate][trulens.core.Feedback.aggregate].
    """

    ZIP = "zip"
    """Match argument values per position in produced values.

    Example:
        If the selector for `arg1` generates values `0, 1, 2` and one for `arg2`
        generates values `"a", "b", "c"`, the feedback function will be called 3
        times with kwargs:

        - `{'arg1': 0, arg2: "a"}`,
        - `{'arg1': 1, arg2: "b"}`,
        - `{'arg1': 2, arg2: "c"}`

    If the quantities of items in the various generators do not match, the
    result will have only as many combinations as the generator with the
    fewest items as per python [zip][zip] (strict mode is not used).

    Note that selectors can use
    [Lens][trulens.core.utils.serial.Lens] `collect()` to name a single (list)
    value instead of multiple values.
    """

    PRODUCT = "product"
    """Evaluate feedback on all combinations of feedback function arguments.

    Example:
        If the selector for `arg1` generates values `0, 1` and the one for
        `arg2` generates values `"a", "b"`, the feedback function will be called
        4 times with kwargs:

        - `{'arg1': 0, arg2: "a"}`,
        - `{'arg1': 0, arg2: "b"}`,
        - `{'arg1': 1, arg2: "a"}`,
        - `{'arg1': 1, arg2: "b"}`

    See [itertools.product][itertools.product] for more.

    Note that selectors can use
    [Lens][trulens.core.utils.serial.Lens] `collect()` to name a single (list)
    value instead of multiple values.
    """


class FeedbackDefinition(
    pyschema_utils.WithClassInfo, serial_utils.SerialModel, Hashable
):
    """Serialized parts of a feedback function.

    The non-serialized parts are in the
    [Feedback][trulens.core.Feedback] class.
    """

    model_config: ClassVar[dict] = dict(arbitrary_types_allowed=True)

    implementation: Optional[
        Union[pyschema_utils.Function, pyschema_utils.Method]
    ] = None
    """Implementation serialization."""

    aggregator: Optional[
        Union[pyschema_utils.Function, pyschema_utils.Method]
    ] = None
    """Aggregator method serialization."""

    examples: Optional[List[Tuple]] = None
    """User supplied examples for this feedback function."""

    criteria: Optional[str] = None
    """Criteria for the feedback function."""

    combinations: Optional[FeedbackCombinations] = FeedbackCombinations.PRODUCT
    """Mode of combining selected values to produce arguments to each feedback
    function call."""

    feedback_definition_id: types_schema.FeedbackDefinitionID
    """Id, if not given, uniquely determined from content."""

    if_exists: Optional[serial_utils.Lens] = None
    """Only execute the feedback function if the following selector names
    something that exists in a record/app.

    Can use this to evaluate conditionally on presence of some calls, for
    example. Feedbacks skipped this way will have a status of
    [FeedbackResultStatus.SKIPPED][trulens.core.schema.feedback.FeedbackResultStatus.SKIPPED].
    """

    if_missing: FeedbackOnMissingParameters = FeedbackOnMissingParameters.ERROR
    """How to handle missing parameters in feedback function calls."""

    run_location: Optional[FeedbackRunLocation]
    """Where the feedback evaluation takes place (e.g. locally, at a Snowflake server, etc)."""

    selectors: Dict[str, serial_utils.Lens]
    """Selectors; pointers into [Records][trulens.core.schema.record.Record] of where
    to get arguments for `imp`."""

    supplied_name: Optional[str] = None
    """An optional name. Only will affect displayed tables."""

    higher_is_better: Optional[bool] = None
    """Feedback result magnitude interpretation."""

    def __init__(
        self,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        implementation: Optional[
            Union[pyschema_utils.Function, pyschema_utils.Method]
        ] = None,
        aggregator: Optional[
            Union[pyschema_utils.Function, pyschema_utils.Method]
        ] = None,
        examples: Optional[List[Tuple]] = None,
        criteria: Optional[str] = None,
        if_exists: Optional[serial_utils.Lens] = None,
        if_missing: FeedbackOnMissingParameters = FeedbackOnMissingParameters.ERROR,
        selectors: Optional[Dict[str, serial_utils.Lens]] = None,
        name: Optional[str] = None,
        higher_is_better: Optional[bool] = None,
        run_location: Optional[FeedbackRunLocation] = None,
        **kwargs,
    ):
        selectors = selectors or {}

        if name is not None:
            kwargs["supplied_name"] = name

        super().__init__(
            feedback_definition_id="temporary",
            implementation=implementation,
            aggregator=aggregator,
            examples=examples,
            criteria=criteria,
            selectors=selectors,
            if_exists=if_exists,
            if_missing=if_missing,
            run_location=run_location,
            **kwargs,
        )

        # By default, higher score is better
        if higher_is_better is None:
            self.higher_is_better = True
        else:
            self.higher_is_better = higher_is_better

        if feedback_definition_id is None:
            if implementation is not None:
                feedback_definition_id = json_utils.obj_id_of_obj(
                    self.model_dump(), prefix="feedback_definition"
                )
            else:
                feedback_definition_id = "anonymous_feedback_definition"

        self.feedback_definition_id = feedback_definition_id

    def __repr__(self):
        return f"FeedbackDefinition({self.name},\n\tselectors={self.selectors},\n\tif_exists={self.if_exists}\n)"

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return hash(self.feedback_definition_id)

    @property
    def name(self) -> str:
        """Name of the feedback function.

        Derived from the name of the serialized implementation function if name
        was not provided.
        """

        if self.supplied_name is not None:
            return self.supplied_name

        if self.implementation is None:
            raise RuntimeError("This feedback function has no implementation.")

        return self.implementation.name


# HACK013: Need these if using __future__.annotations .
FeedbackResult.model_rebuild()
FeedbackCall.model_rebuild()
FeedbackDefinition.model_rebuild()
